static int trace;
/* QVM86 virtualisation module for qemu

   Copyright (C) 2005 CodeSourcery, LLC.
   Written by Paul Brook.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Host kernel side code.
 */
#include "qvm86.h"
//#define QVM86_PROFILE

#ifdef QVM86_PROFILE
static inline uint32_t
readtsc (void)
{
  uint32_t val;
  asm volatile ("rdtsc": "=a" (val) :: "edx");
  return val;
}

uint64_t qvm86_stats[32];
static inline void
profile_timer (int n)
{
  static uint32_t last_timer = 0;
  static uint32_t t;

  t = readtsc();
  if (last_timer)
    {
      qvm86_stats[n] += t - last_timer;
      last_timer = t;
    }
  else
    {
      last_timer = t;
    }
}
#else
#define profile_timer(x) do {} while (0)
#endif

/* TODO: More error checking needed, expecially OOM.  */
/* TODO: Implement some sort of memory management.  Guest pages are mapped
 * in as needed, and only freed when the device is closed.  */
/* TODO: Check for memory leaks.  */
/* TODO: macroize all the places where we do "| 3" to convert addr to pte.  */

#if 0
/* Special host pagetable entries.  */
#define PTEINDEX_RESERVED  0x3fe
#define PTEINDEX_GUEST_PGD 0x3fe
#define PTEINDEX_MON_FREE  0x3ff
#endif

/* Check that the assembly offsets for the monitor_env structure are correct.
   This uses a sneaky trick to force a compiler error when they don't match.
   If you get errors about negative sized arrays, this is why.  */

#define CHECK_OFFSET(name, o, align) extern int __fixme_##name[ \
  ((offsetof(qvm86_monitor_env, name) == o) \
   && (offsetof(qvm86_monitor_env, name) % align == 0)) \
  ? 1 : -1];
#define FIELD_S(s, o, n) CHECK_OFFSET(s##_##n, o, 2)
#define FIELD_L(s, o, n) CHECK_OFFSET(s##_##n, o, 4)
#define FIELD_END(o) extern int __fixme_monitor_env_size[ \
  (sizeof (qvm86_monitor_env) == o ? 1 : -1)];
#include "monitor-env.h"

/* TODO: Disable this for release builds.  */
#define qvm86_assert(cond, msg) \
  do { if (!(cond)) { \
      printk(KERN_ERR "qvm86:assertion failure:" msg "\n");\
      qs->die = 1;\
  }} while (0)

typedef enum 
{
  SHADOW_FAULT_READ,
  SHADOW_FAULT_WRITE,
  SHADOW_FAULT_INTERNAL,
  SHADOW_FAULT_SYSTEM
} shadow_fault_type;

/* Map in a page of guest physical memory.  */
static inline qvm86_mapped_page *
qvm86_map_guest_page (qvm86_state * qs, uint32_t addr)
{
  qvm86_mapped_page *map;
  
  map = &qs->host_physmem_map[addr >> 12];
  if (map->host == NULL)
    {
      host_map_user ((unsigned long) &qs->user_physmem[addr], map);
      qs->monitor_physmem_index[addr >> 22][(addr >> 12) & 0x3ff] = map->phys;
    }
  return map;
}

/* Walk the guest pagetable.  Returns the guest physical address for the
   given virtual address.  */
static uint32_t
qvm86_walk_guest_pgtable_nopae (qvm86_state * qs, uint32_t addr,
                                shadow_fault_type type)
{
  uint32_t guest_pde, guest_pte;
  uint32_t *guest_pdep;
  uint32_t tmp;
  int err;
  qvm86_mapped_page *guest_ptemap;

  addr &= 0xfffff000;

  if (type == SHADOW_FAULT_WRITE)
    err = 0x16;
  else
    err = 0x14;

  tmp = addr >> 22;
  guest_pdep = ((uint32_t *)qs->guest_pgd->host) + tmp;
  guest_pde = *guest_pdep;
  if ((guest_pde & 1) == 0)
    {
      /* Page doesn't exist in the guest either.  */
      return err;
    }
  if ((type != SHADOW_FAULT_SYSTEM && (guest_pde & 4) == 0)
      || (type == SHADOW_FAULT_WRITE && (guest_pde & 2) == 0))
    {
      /* Access violation.  */
      return err | 5;
    }

  /* Set the access bits on the guest pde.  */
  guest_pde |= (1 << 5);
  if (guest_pde & (1 << 7))
    {
      /* 4M page.  Map in a single 4k page.  */
      if (type == SHADOW_FAULT_WRITE)
        guest_pde |= (1 << 6);

      *guest_pdep = guest_pde;
      return (guest_pde & 0xffc00000) | (addr & 0x003ff000);
    }

  *guest_pdep = guest_pde;

  /* Lookup the pte.  */
  tmp = (addr >> 12) & 0x3ff;
  guest_ptemap = qvm86_map_guest_page (qs, guest_pde & 0xfffff000);
  guest_pte = ((uint32_t *)guest_ptemap->host)[tmp];
  if ((guest_pte & 1) == 0)
    {
      /* Page doesn't exist in the guest.  */
      return err;
    }
  if ((type != SHADOW_FAULT_SYSTEM && (guest_pte & 4) == 0)
      || (type == SHADOW_FAULT_WRITE && (guest_pte & 2) == 0))
    {
      /* Access violation.  */
      return err | 5;
    }

  /* Set the access bits on the guest pde.  */
  guest_pte |= (1 << 5);
  if (type == SHADOW_FAULT_WRITE)
    guest_pte |= (1 << 6);
  ((uint32_t *)guest_ptemap->host)[tmp] = guest_pte;
  return guest_pte & 0xfffff000;
}

static int
next_free_gdt_entry (qvm86_state * qs, int n)
{
  while (qs->monitor_gdt[n].high & (1 << 15))
    n++;
  qvm86_assert (n < 8192, "Descriptor overflow");
  return n;
}

static inline void
clear_monitor_gdt_entry(qvm86_state * qs, int n)
{
  qs->monitor_gdt[n].low = 0;
  qs->monitor_gdt[n].high = 0;
}

static void
set_monitor_gdt_entry (qvm86_state * qs, int n, uint32_t base,
			uint32_t limit, uint32_t flags)
{
  if (flags & (1 << 23))
    limit = limit >> 12;
  qs->monitor_gdt[n].low = (base << 16) | (limit & 0xffff);
  qs->monitor_gdt[n].high = (base & 0xff000000) | (limit & 0xf0000)
      | (flags & 0xf0ff00) | ((base >> 16) & 0xff);
}

/* Find the gdt/idt entry for a selector.  May return a pointer to a copy
   of the entry if the descriptor spans a page boundary.  */
static qvm86_gdt_entry *
qvm86_find_selector (qvm86_state * qs, struct kqemu_cpu_state * hostenv,
		     int selector)
{
  uint32_t addr;
  uint32_t physaddr;
  uint32_t hostaddr;
  qvm86_gdt_entry *p;
  static qvm86_gdt_entry copy;

  if (selector < 4)
    return NULL;

  if (selector & 4)
    addr = hostenv->ldt.base;
  else
    addr = hostenv->gdt.base;
  addr += (selector & 0xfff8);

  physaddr = qvm86_walk_guest_pgtable_nopae (qs, addr & 0xfffff000,
                                             SHADOW_FAULT_SYSTEM);
  if (physaddr != 0 && physaddr < 0xfff)
    {
      printk ("Inaccessible selector %x\n", selector);
      qs->die = 1;
    }
  hostaddr = (uint32_t)qs->host_physmem_map[physaddr >> 12].host;
  hostaddr += addr & 0xfff;
  if ((hostaddr & 0xfff) <= 0xff8)
    p = (qvm86_gdt_entry *) hostaddr;
  else
    {
      /* Desriptor overlaps multiple pages.  */
      memcpy (&copy, (void *)hostaddr, 8 - (hostaddr & 7));
      physaddr = qvm86_walk_guest_pgtable_nopae (qs,
                  (addr & 0xfffff000) + 0x1000, SHADOW_FAULT_SYSTEM);
      if (physaddr != 0 && physaddr < 0xfff)
        {
          printk ("Inaccessible selector %x\n", selector);
          qs->die = 1;
        }
      hostaddr = (uint32_t)qs->host_physmem_map[physaddr >> 12].host;
      memcpy (((char *)&copy) + (addr & 7), (void *)hostaddr, 8 - (addr & 7));
      p = &copy;
    }

  return p;
}

static void
qvm86_verify_selector (qvm86_state * qs, struct kqemu_cpu_state * hostenv,
		       int i, qvm86_gdt_entry *p, int pass)
{
  uint32_t limit;
  uint32_t selector;

  selector = hostenv->segs[i].selector;
  if (selector < 4)
    return;

  limit = ((p->low & 0xffff) | (p->high & 0xf0000));
  if (p->high & (1 << 23))
    limit = (limit << 12) | 0xfff;

  if ((hostenv->segs[i].flags &~ 0x100) != (p->high &~ 0x100)
      || hostenv->segs[i].base != ((p->low >> 16) | ((p->high & 0xff) << 16)
			      | (p->high & 0xff000000))
			  
      ||hostenv->segs[i].limit != limit)
    {
      /* This shouldn't really happen.  Interrupts are enabled,
	 so an interrupt could cause the selector to be reloaded.  */
      printk ("Selector mismatch %d %x pass %d\n", i, selector, pass);
      printk ("%08x %08x %08x\n",
	      hostenv->segs[i].flags &~ 0x100,
	      hostenv->segs[i].base,
	      hostenv->segs[i].limit);
      printk ("%08x %08x %08x\n", p->high &~ 0x100,
	  ((p->low >> 16) | ((p->high & 0xff) << 16) | (p->high & 0xff000000)),
	  limit);
      qs->die = 1;
    }
}

/* Clearing the whole gdt is expensive.  Only a handful of descriptors are
   actually used, so just clear those explicitly.  */
static void
qvm86_clear_gdt_entries (qvm86_state *qs, struct kqemu_cpu_state * hostenv)
{
  int i;
  struct kqemu_segment_cache *s;

  for (i = 0; i < 6; i++)
    {
      s = &hostenv->segs[i];
      if ((s->selector & 4) == 0)
	clear_monitor_gdt_entry (qs, s->selector >> 3);
    }
  s = &hostenv->ldt;
  clear_monitor_gdt_entry (qs, s->selector >> 3);
  s = &hostenv->tr;
  clear_monitor_gdt_entry (qs, s->selector >> 3);

  clear_monitor_gdt_entry (qs, qs->monitor_env->mon_cs >> 3);
  clear_monitor_gdt_entry (qs, qs->monitor_env->mon_ss >> 3);
  clear_monitor_gdt_entry (qs, qs->monitor_env->mon_switch_cs >> 3);
  clear_monitor_gdt_entry (qs, qs->monitor_env->mon_switch_ds >> 3);
  clear_monitor_gdt_entry (qs, qs->monitor_env->mon_linmap >> 3);
}

/* Copy gdt entries used by the guest.  */
static void
qvm86_copy_gdt_entries (qvm86_state * qs, struct kqemu_cpu_state * hostenv)
{
  int i;
  struct kqemu_segment_cache *s;

  /* Copy the gdt entries for the active segments.  */
  for (i = 0; i < 6; i++)
    {
      s = &hostenv->segs[i];
      if ((s->selector & 4) == 0)
	set_monitor_gdt_entry (qs, s->selector >> 3, s->base, s->limit,
				s->flags);
    }
  s = &hostenv->ldt;
  set_monitor_gdt_entry (qs, s->selector >> 3, s->base, s->limit,
			  s->flags);
  qs->monitor_env->mon_ldt = s->selector;
  s = &hostenv->tr;
  /* We use the user-specified selector value, but point it at a different
     location.  */
  set_monitor_gdt_entry (qs, s->selector >> 3,
			 qs->monitor_base + qs->maddr_tss, 104,
			 s->flags);
  qs->monitor_env->mon_tr = s->selector;
  /* TODO: If a user process modifies a gdt entry and reloads the
     associated segment descriptor then virtualization will fail.
     Hopefully no real OS will be dumb enough to let user processes
     write to the gdt.  The foolproof solution is to trap writes to
     the guest gdt.  */
}

/* Create dummy gdt entries pointing at the monitor task.
   The monitor code and data segment offsets point to the start of the
   monitor space.  This makes relocating the monitor space easy.  */
static void
qvm86_setup_monitor_gdt (qvm86_state * qs)
{
  int n;

  qs->monitor_env->mon_gdt_base = qs->monitor_base + qs->maddr_gdt;

  /* Code segment.  */
  n = next_free_gdt_entry (qs, 1);
  qs->monitor_env->mon_cs = n << 3;
  set_monitor_gdt_entry (qs, n, qs->monitor_base, 0x003fffff, 0x00c09a00);

  /* Data segment.  We will also use this for the stack.  */
  n = next_free_gdt_entry (qs, n);
  qs->monitor_env->mon_ss = n << 3;
  set_monitor_gdt_entry (qs, n, qs->monitor_base, 0x003fffff, 0x00c09200);

  /* Switch code segment.  */
  n = next_free_gdt_entry (qs, 1);
  qs->monitor_env->mon_switch_cs = n << 3;
  set_monitor_gdt_entry (qs, n, qs->monitor_env->host_env, 0x003fffff,
			0x00c09a00);

  /* Switch data segment.  */
  n = next_free_gdt_entry (qs, 1);
  qs->monitor_env->mon_switch_ds = n << 3;
  set_monitor_gdt_entry (qs, n, qs->monitor_env->host_env, 0x003fffff,
			0x00c09200);

  /* Data segment with 1:1 linear mapping.  Used to access host gdt.  */
  n = next_free_gdt_entry (qs, n);
  qs->monitor_env->mon_linmap = n << 3;
  set_monitor_gdt_entry (qs, n, 0, 0xffffffff, 0x00c09200);
}

/* Fill in the relocatable idt fields.  */
static void
qvm86_setup_monitor_idt (qvm86_state * qs)
{
  qvm86_gdt_entry * p;
  int i;

  if (qs->last_mon_cs == qs->monitor_env->mon_cs)
    return;
  qs->last_mon_cs = qs->monitor_env->mon_cs;
  qs->monitor_env->mon_idt_base = qs->monitor_base + qs->maddr_idt;

  p = qs->monitor_idt;
  for (i = 0; i < 256; i++, p++)
    p->low = (p->low & 0xffff) | (((uint32_t) qs->monitor_env->mon_cs) << 16);
}

/* Create a monitor idt stub.  */
static void
setup_idt_stub (qvm86_state * qs, int i, int type)
{
  qvm86_stub *stub;
  uint32_t maddr_stub;

  maddr_stub = qs->maddr_stubs + i * sizeof (qvm86_stub);

  qs->monitor_idt[i].high = (maddr_stub & 0xffff0000) | 0x8e00;
  /* The segment selector will be filled in later.  */
  qs->monitor_idt[i].low = maddr_stub & 0xffff;
  stub = &qs->monitor_stubs[i];
  switch (type)
    {
    case IDT_INT:
      stub->m2.pushla = 0x68;
      stub->m2.dummy  = 0;
      stub->m2.pushlb = 0x68;
      stub->m2.vector = i;
      stub->m2.jmp    = 0xe9;
      stub->m2.reloc  = qs->maddr_handleint -
	(maddr_stub + sizeof (idt_method2_t));
      break;

    case IDT_EFAULT:
      stub->m1.pushl  = 0x68;
      stub->m1.vector = i;
      stub->m1.jmp    = 0xe9;
      stub->m1.reloc  = qs->maddr_handlefault -
	(maddr_stub + sizeof (idt_method1_t));
      break;

    case IDT_FAULT:
      stub->m2.pushla = 0x68;
      stub->m2.dummy  = 0;
      stub->m2.pushlb = 0x68;
      stub->m2.vector = i;
      stub->m2.jmp    = 0xe9;
      stub->m2.reloc  = qs->maddr_handlefault -
	(maddr_stub + sizeof (idt_method2_t));
      break;

    default:
      printk (KERN_ERR "Invalid ITD entry\n");
      qs->die = 1;
      break;
    }
}

/* Map an pages into monitor space.  PTR is the host virtual address.
   Returns the offset from the monitor base.  */
static uint32_t
map_monitor_pages_pae (qvm86_state * qs, void * ptr, int npages)
{
  uint64_t *pte;
  uint64_t *pagetable;
  uint64_t physaddr;
  int i;

  /* TODO: Check for out-of memory (Should never happen).  */
  pagetable = (uint64_t *) qs->monitor_pagetable;
  /* Look for a free area in the monitor address space.  */
  pte = pagetable;
  for (;;)
    {
      while (*pte & 1)
	pte++;
      for (i = 1; i < npages; i++)
	{
	  if (pte[i] & 1)
	    break;
	}
      if (i == npages)
	break;
      pte += i;
    }
  for (i = 0; i < npages; i++)
    {
      physaddr = host_phys_addr (ptr);
      pte[i] = physaddr | 0x3;
      ptr = (char *)ptr + 4096;
    }
  return (pte - pagetable) << 12;
}

static uint32_t
map_monitor_pages_nopae (qvm86_state * qs, void * ptr, int npages)
{
  uint32_t *pte;
  uint32_t *pagetable;
  uint32_t physaddr;
  int i;

  /* TODO: Check for out-of memory (Should never happen).  */
  pagetable = qs->monitor_pagetable;
  pte = pagetable;
  /* Look for a free area in the monitor address space.  */
  for (;;)
    {
      while (*pte & 1)
	pte++;
      for (i = 1; i < npages; i++)
	{
	  if (pte[i] & 1)
	    break;
	}
      if (i == npages)
	break;
      pte += i;
    }
  for (i = 0; i < npages; i++)
    {
      physaddr = host_phys_addr (ptr);
      pte[i] = physaddr | 0x3;
      ptr = (char *)ptr + 4096;
    }
  return (pte - pagetable) << 12;
}

/* Map an pages into monitor space.  PTR is the host virtual address.
   Returns the offset from the monitor base.  */
static uint32_t
map_monitor_pages (qvm86_state * qs, void * ptr, int npages)
{
  if (qvm86_host_pae)
    return map_monitor_pages_pae (qs, ptr, npages);
  else
    return map_monitor_pages_nopae (qs, ptr, npages);
}

/* Create the monitor interrupt descriptor table and handler stubs.  */
static void
qvm86_create_stubs (qvm86_state * qs)
{
  int i;

  i = (sizeof (qvm86_stub) * 256 + 4095) >> 12;
  qs->monitor_stubs = host_alloc_pages (i);
  qs->maddr_stubs = map_monitor_pages (qs, qs->monitor_stubs, i);

  setup_idt_stub (qs, 0,  IDT_FAULT);
  setup_idt_stub (qs, 2,  IDT_FAULT);
  setup_idt_stub (qs, 2,  IDT_INT); 
  setup_idt_stub (qs, 3,  IDT_FAULT);
  setup_idt_stub (qs, 4,  IDT_FAULT);
  setup_idt_stub (qs, 5,  IDT_FAULT);
  setup_idt_stub (qs, 6,  IDT_FAULT);
  setup_idt_stub (qs, 7,  IDT_FAULT);
  setup_idt_stub (qs, 8,  IDT_EFAULT);
  setup_idt_stub (qs, 9,  IDT_FAULT);
  setup_idt_stub (qs, 10, IDT_EFAULT);
  setup_idt_stub (qs, 11, IDT_EFAULT);
  setup_idt_stub (qs, 12, IDT_EFAULT);
  setup_idt_stub (qs, 13, IDT_EFAULT);
  setup_idt_stub (qs, 14, IDT_EFAULT);
  setup_idt_stub (qs, 15, IDT_FAULT);
  setup_idt_stub (qs, 16, IDT_FAULT);
  setup_idt_stub (qs, 17, IDT_EFAULT);
  setup_idt_stub (qs, 18, IDT_FAULT);
  setup_idt_stub (qs, 19, IDT_FAULT);
  for (i = 20; i < 32; i++)
    setup_idt_stub (qs, i, IDT_FAULT);
  for (i = 32; i < 256; i++)
    setup_idt_stub (qs, i, IDT_INT);
}

/* Set monitor control register values.  */
static void
qvm86_setup_monitor_vm (qvm86_state * qs)
{
  /* TODO: Get some bits from environment (#NE, #TS?).  */
  /* TODO: Make floating point work.  */
  qs->monitor_env->mon_cr0 = 0x8000003b;
  qs->monitor_env->mon_cr3 = qs->monitor_cr3;
  /* TODO: cr4 values?  */
  if (qvm86_host_pae)
    qs->monitor_env->mon_cr4 = 0x30;
  else
    qs->monitor_env->mon_cr4 = 0;
}

/* Allocate the monitor gdt and idt.  */
static void
qvm86_create_monitor_gdt (qvm86_state * qs)
{
  /* The descriptor table can contain at most 8192 entries (=64kbytes).  */
  qs->monitor_gdt = host_alloc_pages (16);
  qs->maddr_gdt = map_monitor_pages (qs, qs->monitor_gdt, 16);
  qs->monitor_env->mon_gdt_limit = 0xffff;

  qs->monitor_idt = host_alloc_pages (1);
  qs->maddr_idt = map_monitor_pages (qs, qs->monitor_idt, 1);
  qs->monitor_env->mon_idt_limit = 0xfff;
}

/* Setup the monitor stack for entry into guest code.  */
static void
qvm86_setup_monitor_stack (qvm86_state * qs, struct kqemu_cpu_state * cpu)
{
  uint32_t *sp;

  sp = qs->monitor_stack;
  
  /* The monitor stack contains an interrupt frame (as if from the guest
     code) plus the saved monitor state.  This must match the code in
     switch.S.  */
  *(--sp) = cpu->segs[2].selector; /* ss */
  *(--sp) = cpu->regs[4]; /* esp */
  *(--sp) = cpu->eflags; /* eflags */
  *(--sp) = cpu->segs[1].selector; /* cs */
  *(--sp) = cpu->eip; /* eip */
  *(--sp) = qs->maddr_start_guest; /* Monitor return address.  */
  *(--sp) = 0x00000002; /* elfags */
  *(--sp) = cpu->regs[0]; /* eax */
  *(--sp) = cpu->regs[1]; /* ecx */
  *(--sp) = cpu->regs[2]; /* edx */
  *(--sp) = cpu->regs[3]; /* ebx */
  *(--sp) = 0; /* esp (ignored) */
  *(--sp) = cpu->regs[5]; /* ebp */
  *(--sp) = cpu->regs[6]; /* esi */
  *(--sp) = cpu->regs[7]; /* edi */
  *(--sp) = cpu->segs[3].selector; /* ds */
  *(--sp) = cpu->segs[0].selector; /* es */
  *(--sp) = cpu->segs[4].selector; /* fs */
  *(--sp) = cpu->segs[5].selector; /* gs */

  /* Point the monitor stack pointer at this data.  */
  qs->monitor_env->mon_esp = qs->maddr_stack - (qs->monitor_stack - sp) * 4;

  /* Setup the interrupt handler stack.  */
  qs->monitor_tss->ss0 = qs->monitor_env->mon_ss;
  qs->monitor_tss->esp0 = qs->maddr_stack - 0x100;
}

#if 0
/* Read a value from guest physical memory.  */
static uint32_t
guest_read_phys (qvm86_state * qs, uint32_t addr)
{
  uint32_t tmp;

  __get_user (tmp, &qs->l1_phys_map[addr >> 22]);
  __get_user (tmp, (uint32_t *)tmp + ((addr >> 12) & 0x3ff));
  if (tmp & 0x3ff)
    {
      printk ("qvm86: Unexpected iomapped page\n");
      qs->die = 1;
      return 0;
    }
  if (tmp != (addr & 0xfffff000))
    {
      printk (KERN_ERR "qvm86: Nonlinear memory mapping: %08x != %08x\n",
	      tmp, addr & 0xfffff000);
      qs->die = 1;
      return 0;
    }
  __get_user (tmp, (uint32_t *) &qs->user_physmem [addr]);
  return tmp;
}
#endif

/* Add a page to the shadow pagetable.  32-bit verson.  */
static int
qvm86_shadow_page_nopae (qvm86_state * qs, uint32_t addr,
                         shadow_fault_type type, qvm86_mapped_page *map)
{
  uint32_t tmp;
  qvm86_shadow_pdmap_nopae *shadow_pdmap;
  uint32_t *shadow_pdep;
  uint32_t shadow_pte;


  /* Setup the shadow PDE.  */
  tmp = (addr >> 22) & 0x3ff;
  shadow_pdmap = &qs->shadow_pdmap_nopae[tmp];
  shadow_pdep = &qs->shadow_pgd_nopae[tmp];
  if ((*shadow_pdep & 1) == 0)
    {
      /* Add this pde to the shadow pagetable.  */
      if (shadow_pdmap->virt == NULL)
	{
	  /* Allocate a new shadow pde.  */
	  shadow_pdmap->virt = (uint32_t *) host_alloc_pages (1);
	  shadow_pdmap->phys = host_phys_addr (shadow_pdmap->virt);
	}
      else
	clear_page (shadow_pdmap->virt);

      /* Add the pagetable to the shadow page directory.  */
      *shadow_pdep = shadow_pdmap->phys | 7;
    }

  /* Setup the shadow PTE.  */
  tmp = (addr >> 12) & 0x3ff;
  shadow_pte = shadow_pdmap->virt[tmp];
  if (shadow_pte & 1)
    {
      /* Page is already mapped.  */
      if ((shadow_pte & 0xfffff000) != map->phys)
        {
          printk ("Page mapping mismatch %08x != %08x @ %08x",
                  shadow_pte, (uint32_t)map->phys, addr);
          qs->die = 1;
          return 0x20;
        }
      if (type == SHADOW_FAULT_INTERNAL || type == SHADOW_FAULT_SYSTEM)
	{
          /* Internal faults may be generated even when the page is already
             shadowed.  */
	  return 0;
	}
      else if (type != SHADOW_FAULT_WRITE)
	{
	  printk ("Page already shadowed %08x@%08x %d\n",
                  addr, shadow_pte, type);
	  qs->die = 1;
	  return 0x20;
	}
      /* Make the page read/write.  */
      shadow_pte |= 2;
    }
  else
    {
      /* Add the page to the pagetable.  */
      shadow_pte = map->phys;
      if (type == SHADOW_FAULT_WRITE)
        shadow_pte |= 7;
      else
        shadow_pte |= 5;
    }

  shadow_pdmap->virt[tmp] = shadow_pte;
  return 0;
}

/* Add a page to the shadow pagetable.  PAE verson.  */
static int
qvm86_shadow_page_pae (qvm86_state * qs, uint32_t addr,
                       shadow_fault_type type, qvm86_mapped_page *map)
{
  uint32_t tmp;
  qvm86_shadow_pdmap_pae *shadow_pdmap;
  uint64_t *shadow_pdp;
  uint64_t shadow_pte;

  /* Find the relevant PGD.  */
  tmp = addr >> 30;
  shadow_pdp = qs->shadow_pgd_pae[tmp];
  tmp = (addr >> 21) & 0x1ff;
  shadow_pdp += tmp;

  /* Setup the shadow PDE.  */
  tmp = (addr >> 21) & 0x7ff;
  shadow_pdmap = &qs->shadow_pdmap_pae[tmp];
  if ((*shadow_pdp & 1) == 0)
    {
      /* Add this pde to the shadow pagetable.  */
      if (shadow_pdmap->virt == NULL)
	{
	  /* Allocate a new shadow pagetable.  */
	  shadow_pdmap->virt = (uint64_t *) host_alloc_pages (1);
	  shadow_pdmap->phys = host_phys_addr (shadow_pdmap->virt);
	}
      else
	clear_page (shadow_pdmap->virt);

      /* Add the pagetable to the shadow page directory.  */
      *shadow_pdp = shadow_pdmap->phys | 7;
    }

  /* Setup the shadow PTE.  */
  tmp = (addr >> 12) & 0x1ff;
  shadow_pte = shadow_pdmap->virt[tmp];
  if (shadow_pte & 1)
    {
      /* Page is already mapped.  */
      if ((shadow_pte & 0xfffff000) != map->phys)
        {
          printk ("Page mapping mismatch %08llx != %08llx @ %08x",
                  shadow_pte, map->phys, addr);
          qs->die = 1;
          return 0x20;
        }
      if (type == SHADOW_FAULT_INTERNAL || type == SHADOW_FAULT_SYSTEM)
	{
          /* Internal faults may be generated even when the page is already
             shadowed.  */
	  return 0;
	}
      else if (type != SHADOW_FAULT_WRITE)
	{
	  printk ("Page already shadowed %08x@%08llx %d\n",
                  addr, shadow_pte, type);
	  qs->die = 1;
	  return 0x20;
	}
      /* Make the page read/write.  */
      shadow_pte |= 2;
    }
  else
    {
      /* Add the page to the pagetable.  */
      shadow_pte = map->phys;
      if (type == SHADOW_FAULT_WRITE)
        shadow_pte |= 7;
      else
        shadow_pte |= 5;
    }

  shadow_pdmap->virt[tmp] = shadow_pte;
  return 0;
}


/* Handle a guest fault.  Returns zero if the page was successfully loaded
   into the shadow pagetable.  Otherwise the low four bits of the return value
   indicate the error code that a real fault would have caused.  A return
   value of 0x20 indicates we were unable to handle the fault.  */
   
static int
qvm86_shadow_fault (qvm86_state * qs, uint32_t addr, shadow_fault_type type)
{
  uint32_t guest_phys;
  uint32_t tmp;
  qvm86_mapped_page *map;

  guest_phys = qvm86_walk_guest_pgtable_nopae (qs, addr, type);

  /* Return fault.  */
  if (guest_phys && guest_phys < 0xfff)
    return guest_phys;

  /* Lookup this page in the emulator physical map.  */
  __get_user (tmp, (uint32_t *)&qs->l1_phys_map[guest_phys >> 22]);
  __get_user (tmp, (uint32_t *)tmp + ((guest_phys >> 12) & 0x3ff));
  if (tmp & 0x3ff)
    {
printk ("iomapped page\n");
      /* This is an IO mapped page.  */
      return 0x20;
    }
  if (tmp != (guest_phys & 0xfffff000))
    {
      printk (KERN_ERR "Nonlinear memory mapping: %08x != %08x\n",
	      tmp, guest_phys);
      guest_phys = tmp & 0xfffff00;
    }

  /* Map the page into memory.  */
  map = qvm86_map_guest_page (qs, guest_phys);

  /* Shadow the page.  */
  if (qvm86_host_pae)
    return qvm86_shadow_page_pae (qs, addr, type, map);
  else
    return qvm86_shadow_page_nopae (qs, addr, type, map);
}

/* Flush a list of pages.  Returns nonzero is a full pagetable flush is
   required.  PAE version.  */
static int
qvm86_flush_list_pae(qvm86_state * qs, int num, unsigned long *list)
{
  int i;
  int n_pte;
  uint32_t addr;
  uint64_t shadow_pde;
  uint32_t guest_pde;

  for (i = 0; i < num; i++)
    {
      __get_user (addr, &list[i]);
      shadow_pde = qs->shadow_pgd_pae[addr >> 30][(addr >> 21) & 0x1ff];
      if (!shadow_pde & 1)
        continue;

      /* ??? This will need fixing for PAE guests.  */
      guest_pde = ((uint32_t *)qs->guest_pgd->host)[addr >> 22];
      /* ??? This looks bogus.  */
      if (((guest_pde ^ shadow_pde) & 0xfffff000) != 0)
        {
          /* It appears that Windows 2000 assumes that invalidating
             a TLB entry with a changed PDE will in fact invalidate
             the whole TLB.  */
          return 1;
        }
      n_pte = (addr >> 12) & 0x1ff;
      qs->shadow_pdmap_pae[addr >> 21].virt[n_pte] = 0;
    }
  return 0;
}

/* Flush a list of pages.  Returns nonzero is a full pagetable flush is
   required.  Non-PAE version.  */
static int
qvm86_flush_list_nopae(qvm86_state * qs, int num, unsigned long *list)
{
  int i;
  int n_pte;
  uint32_t addr;
  uint32_t guest_pde;

  for (i = 0; i < num; i++)
    {
      __get_user (addr, &list[i]);
      if (!qs->shadow_pgd_nopae[addr >> 22])
        continue;

      /* ??? This will need fixing for PAE guests.  */
      guest_pde = ((uint32_t *)qs->guest_pgd->host)[addr >> 22];
      /* ??? This looks bogus.  */
      if (((guest_pde ^ qs->shadow_pgd_nopae[addr >> 22]) & 0xfffff000) != 0)
        {
          /* It appears that Windows 2000 assumes that invalidating
             a TLB entry with a changed PDE will in fact invalidate
             the whole TLB.  */
          return 1;
        }
      n_pte = (addr >> 12) & 0x3ff;
      qs->shadow_pdmap_nopae[addr >> 22].virt[n_pte] = 0;
    }
  return 0;
}

/* Clear invalidated shadow pagetable entries.  */
static void
qvm86_invalidate_pagetables (qvm86_state * qs, struct kqemu_cpu_state * hostenv)
{
  int num;

  num = hostenv->nb_pages_to_flush;
  if (num && num < KQEMU_FLUSH_ALL)
    {
      if (qvm86_host_pae)
        num = qvm86_flush_list_pae (qs, num, qs->pages_to_flush);
      else
        num = qvm86_flush_list_nopae (qs, num, qs->pages_to_flush);
    }

  if (num)
    {
      if (qvm86_host_pae)
        {
          int i;

          for (i = 0; i < 4; i++)
            clear_page (qs->shadow_pgd_pae[i]);
        }
      else
        clear_page (qs->shadow_pgd_nopae);
    }
}

/* Set a host page directory entry.  ADDR is the address of the pagetable.
   n is the top 10 bits of the address to map.  This is a bit of a hack,
   and should probably go away.  */
static void
qvm86_set_monitor_pde(qvm86_state * qs, int n, host_physaddr_t addr)
{
  if (qvm86_host_pae)
    qs->shadow_pgd_pae[n >> 8][(n << 1) & 0x1fe] = addr | 3;
  else
    qs->shadow_pgd_nopae[n] = addr | 3;
}

/* Allocate areas of guest VM space for the monitor.  */
static void
qvm86_setup_monitor_pagetables (qvm86_state * qs)
{
  int i;
  uint32_t pde;
  int n;

  n = 0;
  /* We can use any entry which is absent or supervisor privilege.  */
  for (i = 1023; i >= 0; i--)
    {
      /* ??? Needs fixing for PAE guest.  */
      pde = ((uint32_t *)qs->guest_pgd->host)[i];
      if ((pde & 5) != 5)
	{
	  /* Found a pde that can be used.  */
	  if (n == 0)
	    {
	      qs->monitor_base = i << 22;
              qvm86_set_monitor_pde(qs, i, qs->monitor_pagetable_physaddr);
	    }
#if 0
	  else if (n == 1)
	    {
	      qs->monitor_env->mon_physmem_map = i << 22;
	      qs->shadow_pgd[i] = qs->monitor_physmem_pte_physaddr | 3;
	    }
	  else if (n == 2)
	    {
	      qs->monitor_env->mon_shadow_pte = i << 22;
	      qs->shadow_pgd[i] = qs->monitor_shadow_pte_physaddr | 3;
	    }
#endif
	  else
	    break;
	  n++;
	}
    }
  if (i < 0)
    {
      printk (KERN_ERR "qvm86: Unable to allocate monitor address space\n");
      qs->die = 1;
      return;
    }
}

static inline void
qvm86_unmap_monitor_pagetable (qvm86_state * qs, uint32_t addr)
{
  if (qvm86_host_pae)
    qs->shadow_pgd_pae[addr >> 30][(addr >> 21) & 0x1ff] = 0;
  else
    qs->shadow_pgd_nopae[addr >> 22] = 0;
}

static void
qvm86_unmap_monitor_pagetables (qvm86_state * qs)
{
  qvm86_unmap_monitor_pagetable (qs, qs->monitor_base);
#if 0
  qvm86_unmap_monitor_pagetable (qs->monitor_env->mon_physmem_map);
  qvm86_unmap_monitor_pagetable (qs->monitor_env->mon_shadow_pte);
#endif
}

/* TODO: Put this somewhere host specific.  */
/* TODO: Wheee, evil self-modifying code.  */
#define soft_int(n)                             \
    asm volatile (                              \
        "    movb %b0, __soft_int_vector \n\t"  \
        "    jmp __soft_int_n            \n\t"  \
        "__soft_int_n:                   \n\t"  \
        "    sti                         \n\t"  \
        "    .byte 0xcd                  \n\t"  \
        "__soft_int_vector:              \n\t"  \
        "    .byte 0x00                  \n\t"  \
        :                                       \
        : "r" ((uint8_t) (n) )                  \
        : "memory"                              \
    )

/* Handle and exception from the monitor.  Returns values are:
   0 - continue guest execution.
   1 - Raise exception #14 (page fault)
   2 - Host signal pending, return to userspace.
   >= 32 - Guest software interrupt
 */
static int
qvm86_handle_exception (qvm86_state * qs, int vector)
{
  int err;
  if (vector == 14)
    {
      err = qvm86_shadow_fault (qs, qs->monitor_env->mon_cr2,
				(qs->monitor_env->mon_errcode & 2)
				? SHADOW_FAULT_WRITE : SHADOW_FAULT_READ);
      if (err)
	{
	  if (err & 0x20)
	    return 3;
	  qs->monitor_env->mon_errcode = err;
	  return 1;
	}
      else
	return 0;
    }

  /* All monitor idt entries are DPL0, so INT n instructions generate a
     General Protection Fault.  The monitor contains code to recognise
     these.  */
  if (vector > 0xff)
    return vector;

  if (vector == 2 || vector >= 32)
    {
      /* Hardware interrupt.  Pass it to the host kernel.  */
      /* TODO: Do we need to do anything special with interrupt flags?  */
      /* TODO: Could this cause locking/SMP problems?  Probably not.  */
      soft_int (vector);
      /* Return to userspace if there are signals pending, otherwise
         resume guest execution.  */
      if (host_do_sched(qs))
	return 2;
      else
	return 0;
    }
  /* Some other sort of CPU fault.  */
  /* We currently just punt, and get userspace to execute the instruction.  */
  /* TODO: Some faults are a side-effect of virtualization.  Ideally we
     handle these here.  */
  /* TODO: Some faults are wrong.  For example we give all interrupts a
     DPL of 0, so INT n instructions will generate a General Protection
     Fault.  */
  return 3;
}

/* We have run as much as we can.  Copy the guest state back.  */
static void
qvm86_flush_guest_state (qvm86_state * qs, struct kqemu_cpu_state * hostenv)
{
  uint32_t *sp;
  uint16_t segregs[6];
  int i;
  uint16_t selector;
  uint32_t newflags;
  qvm86_gdt_entry *p;

  /* Guest state is saved on the monitor stack.  Some state is pushed by the
     cpu when the interrupt occurs, the rest is pushed by the monitor code.  */
  i = (qs->maddr_stack - qs->monitor_env->mon_esp) / 4;
  sp = qs->monitor_stack - i;

  segregs[5] = *(sp++); /* gs */
  segregs[4] = *(sp++); /* fs */
  segregs[0] = *(sp++); /* es */
  segregs[3] = *(sp++); /* ds */
  hostenv->regs[7] = *(sp++); /* edi */
  hostenv->regs[6] = *(sp++); /* esi */
  hostenv->regs[5] = *(sp++); /* ebp */
  sp++; /* skip (monitor esp) */
  hostenv->regs[3] = *(sp++); /* ebx */
  hostenv->regs[2] = *(sp++); /* edx */
  hostenv->regs[1] = *(sp++); /* ecx */
  hostenv->regs[0] = *(sp++); /* eax */
  sp++; /* skip (monitor eflags) */
  sp++; /* skip (monitor return address) */
  sp++; /* skip (vector) */
  sp++; /* skip (error code) */
  hostenv->eip = *(sp++); /* eip */
  segregs[1] = *(sp++); /* cs */
  newflags = *(sp++); /* eflags */
  hostenv->regs[4] = *(sp++); /* esp */
  segregs[2] = *(sp++); /* ss */

if (trace){
printk ("After: eip %08x, cr0 %08x\n", hostenv->eip, qs->monitor_env->mon_cr0);
printk ("eax:%08x, ebx:%08x, ecx:%08x, edx:%08x\n", hostenv->regs[0], hostenv->regs[3], hostenv->regs[1], hostenv->regs[2]);
printk ("esi:%08x, edi:%08x, ebp:%08x, esp:%08x\n", hostenv->regs[6], hostenv->regs[7], hostenv->regs[5], hostenv->regs[4]);
}
  hostenv->eflags = (newflags & 0x1ff) | (hostenv->eflags & 0xfffffe00);
  /* Update the cached selector values.  */
  for (i = 0; i < 6; i++)
    {
      selector = segregs[i];

      p = qvm86_find_selector (qs, hostenv, selector);

      if (hostenv->segs[i].selector == selector)
	{
	  qvm86_verify_selector (qs, hostenv, i, p, 1);
	  continue;
	}

      hostenv->segs[i].selector = selector;
      if (selector < 4)
	{
	  hostenv->segs[i].flags = 0;
	  hostenv->segs[i].base = 0;
	  hostenv->segs[i].limit = 0;
	}
      else
	{
	  hostenv->segs[i].flags = p->high;
	  hostenv->segs[i].base = (p->low >> 16) | ((p->high & 0xff) << 16)
				  | (p->high & 0xff000000);
	  hostenv->segs[i].limit = (p->low & 0xffff) | (p->high & 0xf0000);
	  if (p->high & (1 << 23))
	    hostenv->segs[i].limit = (hostenv->segs[i].limit << 12) | 0xfff;
	}
    }
}

/* Upload guest state and transfer control to the monitor.  */
int
qvm86_exec (qvm86_state * qs, struct kqemu_cpu_state * hostenv)
{
  int i;
  int err;
  int retval = KQEMU_RET_ABORT;
  int next_offset;

if (trace) trace--;
if (hostenv->regs[3] == 0x12345678)
{
printk ("Trace\n");
trace = 3;
}
if (trace){
printk ("Before: eip %08x\n", hostenv->eip);
printk ("eax:%08x, ebx:%08x, ecx:%08x, edx:%08x\n", hostenv->regs[0], hostenv->regs[3], hostenv->regs[1], hostenv->regs[2]);
printk ("esi:%08x, edi:%08x, ebp:%08x, esp:%08x\n", hostenv->regs[6], hostenv->regs[7], hostenv->regs[5], hostenv->regs[4]);
}
  /* We currently assume paging is enabled.  */
  if ((hostenv->cr0 & 0x80000000) == 0)
    return KQEMU_RET_SOFTMMU;

  if (qs->guest_cr3 != hostenv->cr3)
    {
      qs->guest_cr3 = hostenv->cr3;
      /* Map the guest page directory in.  */
      qs->guest_pgd = qvm86_map_guest_page (qs, qs->guest_cr3);
#if 0
      /* ??? Does the monitor need the guest pagetable mapped in?  */
      /* Also map it into the monitor.  */
      qs->monitor_pagetable[PTEINDEX_GUEST_PGD] = qs->guest_pgd->phys | 3;
#endif
    }

  qvm86_invalidate_pagetables (qs, hostenv);
  qvm86_setup_monitor_pagetables (qs);

  /* Copy the entries the guest is currently using.  */
  qvm86_copy_gdt_entries (qs, hostenv);

  /* Create gdt segment entries for the monitor.  */
  qvm86_setup_monitor_gdt (qs);

  qvm86_setup_monitor_idt (qs);

  qvm86_setup_monitor_vm (qs);

  qvm86_setup_monitor_stack (qs, hostenv);
  /* We know we will need the page containing the guest code, so fault that in
     now.  */
  qs->monitor_env->mon_errcode = 0;
  i = qvm86_shadow_fault (qs, hostenv->segs[1].base + hostenv->eip,
			  SHADOW_FAULT_INTERNAL);

  if (i && (i & 1) == 0)
    {
      /* Code page not present.  Give up.  */
      qvm86_unmap_monitor_pagetables (qs);
      qvm86_clear_gdt_entries (qs, hostenv);

      if (i == 0x20)
	return KQEMU_RET_SOFTMMU;
      else
	{
	  hostenv->error_code = i & 0xf;
	  hostenv->cr2 = hostenv->segs[1].base + hostenv->eip;
	  return KQEMU_RET_EXCEPTION + 14;
	}
    }
  /* Fault in the gdt.  */
  for (i = 0; i < hostenv->gdt.limit; i += 4096)
    {
      err = qvm86_shadow_fault (qs, hostenv->gdt.base + i,
				SHADOW_FAULT_SYSTEM);
      if (err)
	{
	  printk ("Shadow fault failed on gdt page %d\n", i >> 12);
	  qs->die = 1;
	}
    }
  /* Also fault in the ldt.  */
  if (hostenv->ldt.selector)
    {
      /* TODO: Is it going to cause problems if these get paged out?  */
      for (i = 0; i < hostenv->ldt.limit; i += 4096)
	{
	  err = qvm86_shadow_fault (qs, hostenv->ldt.base + i,
				    SHADOW_FAULT_SYSTEM);
	  if (err)
	    {
	      printk ("Shadow fault failed on gdt page %d\n", i >> 12);
	      qs->die = 1;
	    }
	}
    }

  /* If the cached selector values don't match the values in the descriptor
     tables, we're stuck.  Reloading the values will obviously break, and
     creating fake discriptors will break user code that relods the segment
     register.  */
  for (i = 0; i < 6; i++)
    {
      qvm86_gdt_entry *p;

      p = qvm86_find_selector (qs, hostenv, hostenv->segs[i].selector);
      qvm86_verify_selector (qs, hostenv, i, p, 0);
    }

  next_offset = 0;
  for (;;)
    {
      /* Clear the busy bit in the host task.  */
      qs->monitor_gdt[qs->monitor_env->mon_tr >> 3].high &=
	~(uint32_t) (1 << 9);

      /* Bail out if something went wrong.  */
      if (qs->die)
	return KQEMU_RET_ABORT;

profile_timer(0);
      /* Setup monitor state.  */
      qs->monitor_run (qs->monitor_env);
profile_timer(1);

      i = qvm86_handle_exception (qs, qs->monitor_env->mon_reason);
//if (trace)
//printk ("Exception %d(%d) %08x\n", i, qs->monitor_env->mon_reason, qs->monitor_env->mon_cr2);
      switch (i)
	{
	case 0:
	  break;
	case 1:
	  retval = KQEMU_RET_EXCEPTION + 14;
	  /* Update cr2 with the fault address.  */
	  hostenv->cr2 = qs->monitor_env->mon_cr2;
	  hostenv->error_code = qs->monitor_env->mon_errcode;
	  break;
	case 2:
	  retval = KQEMU_RET_INTR;
	  break;
	case 3:
	  retval = KQEMU_RET_SOFTMMU;
	  break;
	default:
	  retval = (i & 0xff) | KQEMU_RET_INT;
	  next_offset = i >> 8;
	  break;
	}
      if (i)
	break;
    }
#ifdef QVM86_PROFILE
{
static int n = 0;
if (n++ >= 0x1000)
{
/* Soft faults, hard faults, guest faults, hard interrupts, exceptions,
   soft interrupts.  */
for (n = 0; n < 32; n++)
  {
  printk ("%2d:%06x ", n, (uint32_t)(qvm86_stats[n] >> 16));
  qvm86_stats[n] = 0;
  }
printk("\n");
}
}
#endif
  /* Clear temporary data structures.  */
  qvm86_unmap_monitor_pagetables (qs);
  qvm86_clear_gdt_entries (qs, hostenv);
  qvm86_flush_guest_state (qs, hostenv);
  if (next_offset)
    hostenv->next_eip = hostenv->eip + next_offset;

  return retval;
}

/* Create a tss for the monitor.  */
static void
qvm86_create_monitor_tss (qvm86_state * qs)
{
  qs->monitor_tss = (qvm86_tss *) host_alloc_pages (1);
  qs->monitor_tss->iomap = 0x1000;
  qs->maddr_tss = map_monitor_pages (qs, qs->monitor_tss, 1);
}

extern char qvm86_monitor_start[];
extern char qvm86_monitor_end[];

static uint32_t
get_code_maddr (void *p)
{
  return (uint32_t) p - (uint32_t) qvm86_monitor_start;
}

/* Create the shadow page directory tables.  */
static void
qvm86_init_shadow_pgd (qvm86_state * qs)
{
  if (qvm86_host_pae)
    {
      int i;
      uint64_t *shadow_pdpt;
      uint64_t addr;

      qs->shadow_pdmap_pae = (qvm86_shadow_pdmap_pae *)
        host_alloc_zeroed (2048 * sizeof(qvm86_shadow_pdmap_pae));
      shadow_pdpt = (uint64_t *) host_alloc_pages(1);
      /* ??? This must be in the first 4G of physical memory.  */
      addr = host_phys_addr (shadow_pdpt);
      if (addr >> 32 != 0)
        {
          printk ("Monitor CR3 not < 4G");
          qs->die = 1;
        }
      qs->monitor_cr3 = addr;
      for (i = 0; i < 4; i++)
        {
          qs->shadow_pgd_pae[i] = host_alloc_pages (1);
          shadow_pdpt[i] = host_phys_addr (qs->shadow_pgd_pae[i]) | 1;
        }
    }
  else
    {
      qs->shadow_pdmap_nopae = (qvm86_shadow_pdmap_nopae *)
        host_alloc_zeroed (1024 * sizeof(qvm86_shadow_pdmap_nopae));
      qs->shadow_pgd_nopae = host_alloc_pages (1);
      qs->monitor_cr3 = host_phys_addr (qs->shadow_pgd_nopae);
    }
}

/* Create the monitor_env structure and monitor pagetable.  */
static void
qvm86_init_monitor_env (qvm86_state * qs)
{
  int npages;
  uint32_t addr;
  uint32_t switch_size;
  uint32_t monitor_env_addr;
  int n;
  
  /* TODO: This assumes that the size of monitor_env provides sufficient
     alignment guarantees for the code.  It is probably better to
     explicitly put space for monitor_env at the start of switch.S.  */
  /* Copy the monitor code into monitor space.  */
  switch_size = qvm86_monitor_end - qvm86_monitor_start;
  npages = (switch_size + 4095) >> 12;
  /* The shared monitor code is mapped by replacing a single pde.
     This is guaranteed to work if the shared area is a single page.
     If this ever grows larger than a page we will need to play tricks
     to make sure it is never crosses a page directory boundary.
     There is some wiggle room as we only actually need the monitor_env
     structure and the switch code in both address spaces.  */
  qvm86_assert (npages == 1, "Switch data too big");

  qs->monitor_env = (qvm86_monitor_env *) host_alloc_pages (npages);
  memcpy (qs->monitor_env, qvm86_monitor_start, switch_size);

  qvm86_init_shadow_pgd(qs);

  /* ??? Comment is bogus.  */
  /* Map the monitor into monitor space.  We also create pagetable entries
     to map the code and monitor_env structure at the same location as in
     the host. this is used when switching to and from the host.  */
  qs->monitor_pagetable = host_alloc_pages (1);
  qs->monitor_pagetable_physaddr = host_phys_addr (qs->monitor_pagetable);
  /* The monitor_env structure should always be the first page mapped
     into monitor space.  */
  if (map_monitor_pages(qs, qs->monitor_env, 1) != 0)
    {
      printk ("monitor code ended up somewhere funny");
      qs->die = 1;
    }

  monitor_env_addr = (uint32_t) qs->monitor_env;
  if (qvm86_host_pae)
    {
      uint64_t *table;
      uint64_t physaddr;

      table = (uint64_t *) host_alloc_pages (1);
      n = (monitor_env_addr >> 12) & 0x1ff;
      physaddr = host_phys_addr (qs->monitor_env);
      table[n] = physaddr | 0x3;

      physaddr = host_phys_addr (table);
      qs->monitor_env->host_switch_pde_low = physaddr | 0x3;
      qs->monitor_env->host_switch_pde_high = physaddr >> 32;

      table = qs->shadow_pgd_pae[monitor_env_addr >> 30];
      n = (monitor_env_addr >> 21) & 0x1ff;
      qs->monitor_env->host_switch_ptr = (uint32_t) &table[n];
      addr = map_monitor_pages (qs, table, 1);
      qs->monitor_env->mon_switch_ptr = addr + n * 8;
    }
  else
    {
      uint32_t *table;
      uint32_t physaddr;

      table = (uint32_t *) host_alloc_pages (1);
      n = (monitor_env_addr >> 12) & 0x3ff;
      physaddr = host_phys_addr (qs->monitor_env);
      table[n] = physaddr | 0x3;

      physaddr = host_phys_addr (table);
      qs->monitor_env->host_switch_pde_low = physaddr | 0x3;

      n = monitor_env_addr >> 22;
      table = qs->shadow_pgd_nopae;
      qs->monitor_env->host_switch_ptr = (uint32_t) &table[n];
      addr = map_monitor_pages (qs, table, 1);
      qs->monitor_env->mon_switch_ptr = addr + n * 4;
    }

  qs->monitor_env->host_pae = qvm86_host_pae;
  qs->maddr_handleint = get_code_maddr (&qvm86_monitor_inthandler);
  qs->maddr_handlefault = get_code_maddr (&qvm86_monitor_faulthandler);
  qs->maddr_start_guest = get_code_maddr (&qvm86_start_guest);
  qs->monitor_env->mon_eip = get_code_maddr (&qvm86_monitor_entry);
  qs->monitor_env->mon_switch_eip = get_code_maddr (&qvm86_monitor_switch);
  
  qs->monitor_env->host_env = monitor_env_addr;
  qs->monitor_run = (qvm86_monitorproc) (monitor_env_addr
		      + get_code_maddr (&qvm86_monitor_run));
  qs->monitor_env->host_eip = monitor_env_addr
			      + get_code_maddr (&qvm86_monitor_return);

#if 0
  /* The last entry in the monitor pagetable should always be free.  */
  addr = map_monitor_pages (qs, qs->monitor_pagetable, 1);
  qs->monitor_env->mon_free_pte = addr + (PTEINDEX_MON_FREE << 2);
  qs->monitor_env->mon_free_page = PTEINDEX_MON_FREE << 12;

  /* Also allow access to the shadow page directory.  */
  qs->monitor_env->mon_shadow_pgd = map_monitor_pages (qs, qs->shadow_pgd, 1);

  /* Tell the monitor where to find the guest page directory.  */
  qs->monitor_env->mon_guest_pgd = PTEINDEX_GUEST_PGD << 12;
#endif
}

/* Create the monitor stack.  */
static void
qvm86_init_monitor_stack (qvm86_state * qs)
{
  uint32_t *sp;
  sp = (uint32_t *) host_alloc_pages (1);
  qs->monitor_stack = &sp[1024];
  qs->maddr_stack = map_monitor_pages (qs, sp, 1) + 4096;
}

/* Create the host->guest physmem mapping tables.  */
static void
qvm86_init_shadow (qvm86_state * qs)
{
  uint32_t n;
  uint32_t count;
  
  qs->monitor_physmem_index = host_alloc_pages (1);
  qs->monitor_physmem_pte = host_alloc_pages (1);
  qs->monitor_physmem_pte_physaddr
    = host_phys_addr (qs->monitor_physmem_pte);

  count = (qs->user_physmem_size + 0x003fffff) >> 22;
  for (n = 0; n < count; n++)
    {
      qs->monitor_physmem_index[n] = host_alloc_pages (1);
      qs->monitor_physmem_pte[n]
	= host_phys_addr (qs->monitor_physmem_index[n]) | 3;
    }
  qs->monitor_shadow_pte = host_alloc_pages (1);
  qs->monitor_shadow_pte_physaddr = host_phys_addr (qs->monitor_shadow_pte);
}

/* Initialise the guest VM.  */
void
qvm86_init (qvm86_state * qs, struct kqemu_init * init)
{
  printk (KERN_INFO "qvm86: Creating guest VM\n");
  qvm86_init_monitor_env (qs);
  qs->user_physmem = init->ram_base;
  /* TODO: This is a big block, and doesn't need to be mappable memory.  */
  qs->host_physmem_map = (qvm86_mapped_page *)
    host_alloc_zeroed ((init->ram_size >> 12) * sizeof(qvm86_mapped_page));
  qs->user_physmem_size = init->ram_size;
  qs->l1_phys_map = (uint32_t *) init->phys_to_ram_map;
  qs->phys_ram_dirty = init->ram_dirty;
  qs->pages_to_flush = init->pages_to_flush;
  qvm86_init_shadow (qs);
  qvm86_init_monitor_stack (qs);
  qvm86_create_monitor_gdt (qs);
  qvm86_create_stubs (qs);
  qvm86_create_monitor_tss (qs);
}

/* Perform initialisation when device is opened.  */
void
qvm86_init_device (qvm86_state * qs)
{
}

/* Cleanup when device is closed.  */
void
qvm86_cleanup_device (qvm86_state * qs)
{
  int i;

  if (!qs->user_physmem)
    return;

  if (qs->host_physmem_map)
    {
      for (i = 0; i < (qs->user_physmem_size >> 12); i++)
	{
	  if (qs->host_physmem_map[i].host)
	    host_unmap_user (&qs->host_physmem_map[i], 1);
	}
      host_free (qs->host_physmem_map);
    }

  i = (sizeof (qvm86_stub) * 256 + 4095) >> 12;
  host_free_pages (qs->monitor_stubs, i);
  host_free_pages (qs->monitor_gdt, 16);
  host_free_pages (qs->monitor_idt, 1);
  host_free_pages (qs->monitor_pagetable, 1);
  if (qvm86_host_pae)
    {
      if (qs->shadow_pdmap_pae)
        {
          for (i = 0; i < 1024; i++)
            {
              if (qs->shadow_pdmap_pae[i].virt)
                host_free_pages (qs->shadow_pdmap_pae[i].virt, 1);
            }
          host_free (qs->shadow_pdmap_pae);
        }
    }
  else
    {
      if (qs->shadow_pdmap_nopae)
        {
          for (i = 0; i < 1024; i++)
            {
              if (qs->shadow_pdmap_nopae[i].virt)
                host_free_pages (qs->shadow_pdmap_nopae[i].virt, 1);
            }
          host_free (qs->shadow_pdmap_nopae);
        }
    }
  /* ??? host_free_pages (qs->shadow_pgd, 1);*/
  host_free_pages (qs->monitor_env, 1);
  host_free_pages (&qs->monitor_stack[-1024], 1);
  host_free_pages (qs->monitor_tss, 1);
  printk (KERN_INFO "qvm86: Device closed\n");
}
