/* QVM86 virtualisation module for qemu

   Copyright (C) 2005 CodeSourcery, LLC.
   Written by Paul Brook.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Linux kernel interfacing code.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/version.h>
#include <linux/ioctl.h>
#include <linux/smp_lock.h>
#include <linux/version.h>
#include <linux/highmem.h>
#include <linux/miscdevice.h>
#include <asm/atomic.h>
#include <asm/processor.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include "qvm86.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
/* Various bits need newish kernel.  */
#error "Linux kernel 2.4.20 or above required."
#endif

#if defined(THIS_MODULE)
/* Starting with version 2.1.18, the __this_module symbol is present,
   but the THIS_MODULE #define was introduced much later.  */
#define THIS_MODULE (&__this_module)
#endif

#ifndef EXPORT_NO_SYMBOLS
#  define EXPORT_NO_SYMBOLS register_symtab(NULL)
#endif

/* New License scheme. */
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/* Give other tasks a chance to run.  Returns nonzero if this task has
   signals pending.  */
int
host_do_sched (qvm86_state *qs)
{
  if (need_resched())
      schedule();

  return signal_pending(current);
}

void *
host_alloc_zeroed (unsigned long size)
{
  void * ptr;

  ptr = (void *) vmalloc (size);
  if (!ptr)
    return NULL;
  memset (ptr, 0, size);
  return ptr;
}

void *
host_alloc_pages (int count)
{
  char * ptr;

  ptr = (void *) __get_free_pages (GFP_KERNEL, get_order (count << 12));
  if (!ptr)
    return NULL;
  while (count--)
    clear_page (&ptr[count << 12]);
  return ptr;
}

/* Free memory allocated by host_alloc_zeroed.  */
void
host_free (void * ptr)
{
  if (!ptr)
    return;
  vfree (ptr);
}

/* Free pages allocated by host_alloc_pages.  */
void
host_free_pages (void * ptr, int count)
{
  if (!ptr)
    return;
  free_pages ((unsigned long) ptr, get_order (count << 12));
}

host_physaddr_t
host_phys_addr (void * addr)
{
  uint32_t pa;
  pa = virt_to_phys (addr);
  return pa;
}

/* Map user pages into kernel space.  */
void
host_map_user (unsigned long useraddr, qvm86_mapped_page * map)
{
  int ret;
  struct page *page;

  ret = get_user_pages(current,
                       current->mm,
                       (unsigned long) useraddr,
                       1, /* 1 page. */
                       1, /* 'write': intent to write. */
                       0, /* 'force': ? */
                       &page,
                       NULL /* struct vm_area_struct *[] */
                       );
  /* TODO: error handling.  */
  if (ret != 1)
    printk (KERN_ERR "qvm86: host_map_user failed\n");

  map->hostdata = page;
  map->host = kmap (page);
  map->phys = host_phys_addr (map->host);
}

/* Unmap a user page.  */
void
host_unmap_user (qvm86_mapped_page *map, int dirty)
{
  struct page *page;

  page = (struct page *)map->hostdata;
  kunmap (page);
  if (dirty)
    set_page_dirty (page);
  put_page (page);
  /* TODO: This should probably be done elsewhere.  */
  map->host = NULL;
  map->phys = 0;
}

/* Mark a user page as dirty.  */
void
host_mark_dirty (qvm86_mapped_page *map)
{
  struct page *page;

  page = (struct page *) map->hostdata;
  set_page_dirty (page);
}

static int
qvm86_open (struct inode *inode, struct file *filp)
{
  qvm86_state *qs;

  qs = host_alloc_zeroed (sizeof (qvm86_state));
  if (!qs)
    return -ENOMEM;
  filp->private_data = qs;

  qvm86_init_device (qs);

  return 0;
}

static int
qvm86_release (struct inode *inode, struct file *filp)
{
  qvm86_state *qs = (qvm86_state *)filp->private_data;

  qvm86_cleanup_device (qs);
  host_free (qs);
  
  filp->private_data = NULL;
  return 0;
}

/* Handle an ioctl call.  */
static int
qvm86_linux_ioctl (struct inode *inode, struct file *filp,
		   unsigned int cmd, unsigned long arg)
{
  qvm86_state *qs = (qvm86_state *)filp->private_data;
  int ret = 0;

  switch (cmd)
    {
    case KQEMU_EXEC:
      {
	struct kqemu_cpu_state s;
	copy_from_user (&s, (void *) arg, sizeof(s));

	ret = qvm86_exec (qs, &s);
	copy_to_user ((void *) arg, &s, sizeof(s));
      }
      break;

    case KQEMU_INIT:
      {
	struct kqemu_init d;
	copy_from_user (&d, (void *) arg, sizeof (d));
	qvm86_init (qs, &d);
      }
      break;

    case KQEMU_GET_VERSION:
      if (put_user(KQEMU_VERSION, (int *)arg) < 0)
	ret = -EFAULT;
      break;

    default:
      ret = -ENOIOCTLCMD;
      break;
    }
  if (qs->die)
    return KQEMU_RET_ABORT;

  return ret;
}

static struct file_operations qvm86_fops =
{
  owner:    THIS_MODULE,
  ioctl:    qvm86_linux_ioctl,
  open:     qvm86_open,
  release:  qvm86_release,
};

static struct miscdevice qvm86_dev =
{
  .minor      = MISC_DYNAMIC_MINOR,
  .name       = "qvm86",
  .fops       = &qvm86_fops,
};

static int __init
qvm86_module_init (void)
{
  int err;

  printk(KERN_INFO "qvm86: Module loaded\n");
  
  err = misc_register (&qvm86_dev);
  if (err < 0)
    {
      printk(KERN_WARNING "qvm86: Device creation failed\n");
      return err;
    }

  printk (KERN_INFO "qvm86: Created device %d.%d\n",
	  MISC_MAJOR, qvm86_dev.minor);

  return 0;
}
module_init(qvm86_module_init);

static void __exit
qvm86_module_exit (void)
{
  /* Unregister device. */
  misc_deregister (&qvm86_dev);
  printk(KERN_INFO "qvm86: Module removed\n");
}
module_exit(qvm86_module_exit);
