/* QVM86 virtualisation module for qemu

   Copyright (C) 2005 CodeSourcery, LLC.
   Written by Paul Brook.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Header file with declarations for everything.
 */
/* TODO: Split this into qemu-host.h and qemu-monitor.h.  */

#ifdef _WIN32
#include "qvm86-win32.h"
#else
#include "qvm86-linux.h"
#endif

#include "kqemu.h"

#define FIELD_S(s, o, n) uint16_t s##_##n;
#define FIELD_L(s, o, n) uint32_t s##_##n;
#define FIELD_END(o)
typedef struct {
#include "monitor-env.h"
} qvm86_monitor_env;

/* A short description of how qvm86 works.

   There are three VMs:
   - The host VM.  This is the VM maintained by the host OS.  We only access
   this indirectly via host_phys_addr et al.
   - The guest VM.  This is the the emulated CPU/MMU.  
   - The monitor/shadow VM.  This is the VM created by qvm86.  From userspace
   it looks like the guest VM.  The privileged part of this VM a stubs
   that can switch back to the host VM, and an implementation of the
   fast-path of qvm86_shadow_fault.  These stubs are implemented in switch.S.

   The monitor/shadow VM has its own GDT, IDT, TSS and pagetables.  The code
   in switch.S saves the host values and loads the monitor values.  When a
   guest exception occurs it saves the guest state and restores the original
   host values.

   The shadow pagetables start off empty.  We build them on-demand, updating
   the accessed and dirty bits in the guest pagetable as the
   shadow pagetable is populated.  If a shadow pagetable entry (PTE) is
   created in response to a memory read, the shadow PTE will be marked as
   readonly.  This ensures that subsequent writes trigger a fault, which is
   used to update te guest PTE dirty bit.  */

typedef struct
{
  uint32_t low;
  uint32_t high;
} qvm86_gdt_entry;

/* The contents of a TSS segment.  */
typedef struct
{
  uint16_t prev;
  uint16_t pad0;
  uint32_t esp0;
  uint16_t ss0;
  uint16_t pad1;
  uint32_t esp1;
  uint16_t ss1;
  uint16_t pad2;
  uint32_t esp2;
  uint16_t ss2;
  uint16_t pad3;
  uint32_t cr3;
  uint32_t eip;

  uint32_t eflags;
  uint32_t eax;
  uint32_t ecx;
  uint32_t edx;
  uint32_t ebx;
  uint32_t esp;
  uint32_t ebp;
  uint32_t esi;
  uint32_t edi;

  uint16_t es;
  uint16_t pad4;
  uint16_t cs;
  uint16_t pad5;
  uint16_t ss;
  uint16_t pad6;
  uint16_t ds;
  uint16_t pad7;
  uint16_t fs;
  uint16_t pad8;
  uint16_t gs;
  uint16_t pad9;
  uint16_t ldt;
  uint16_t pad10;
  
  uint16_t T:1;
  uint16_t pad11:15;
  uint16_t iomap;
} qvm86_tss;

/* A page of guest memory.  */
typedef struct
{
  void *host;
  
  host_physaddr_t phys;  /* TODO: This field can be removed???.  */
  /* We used to keep track of pte entries for pages so the we could swap them
     back out.  Need to figure out how to do this when a page is referenced
     by multiple pagetable entries.  This happens eg. with the "zero" page
     under linux.  */
  void * hostdata;
} qvm86_mapped_page;

/* A shadow PDE.  */
typedef struct
{
  /* The shadow pagetable.  */
  uint32_t *virt;
  /* The physical address of the shadow pagetable.  */
  uint32_t phys;
} qvm86_shadow_pdmap_nopae;

/* A shadow PDE in PAE mode.  */
typedef struct
{
  /* The physical address of the shadow pagetable.  */
  uint64_t phys;
  /* The shadow pagetable.  */
  uint64_t *virt;
} qvm86_shadow_pdmap_pae;

/* Stub used for trap handlers where the CPU already pushed an error code.  */
typedef struct
{
  uint8_t  pushl;    /* Always 0x68 == pushl            */
  uint32_t vector;   /* Interrupt vector number         */
  uint8_t  jmp;      /* Always 0xe9 == jmp              */
  uint32_t reloc;    /* Relative offset of destination  */
} __attribute__ ((packed)) idt_method1_t;

/* Stub used for trap handlers where the CPU did not push an error code.  */
typedef struct
{
  uint8_t  pushla;   /* Always 0x68 == pushl            */
  uint32_t dummy;    /* Dummy error code                */
  uint8_t  pushlb;   /* Always 0x68 == pushl            */
  uint32_t vector;   /* Interrupt vector number         */
  uint8_t  jmp;      /* Always 0xe9 == jmp              */
  uint32_t reloc;    /* Relative offset of destination  */
} __attribute__ ((packed)) idt_method2_t;

typedef union
{
  idt_method1_t m1;
  idt_method2_t m2;
} qvm86_stub;

#define IDT_INT		0 /* Interrupt.  */
#define IDT_EFAULT	1 /* Fault with error code.  */
#define IDT_FAULT	2 /* Fault without error code.  */

typedef void asmlinkage (*qvm86_monitorproc)(qvm86_monitor_env *);

typedef struct
{
  qvm86_monitor_env * monitor_env;
  qvm86_gdt_entry * monitor_gdt;
  qvm86_gdt_entry * monitor_idt;
  qvm86_stub *monitor_stubs;
  /* Points to the top of the monitor stack.  */
  uint32_t *monitor_stack;
  /* The task state structure for the monitor task.  */
  qvm86_tss * monitor_tss;
  /* The shadow page directory.  */ 
  uint32_t *shadow_pgd_nopae;
  uint64_t *shadow_pgd_pae[4];
  /* The shadow pagetable mappings.  */
  qvm86_shadow_pdmap_nopae *shadow_pdmap_nopae;
  qvm86_shadow_pdmap_pae *shadow_pdmap_pae;
  uint32_t guest_cr3;
  uint32_t monitor_cr3;
  /* The pagetable entries for monitor space.  */
  void *monitor_pagetable;
  host_physaddr_t monitor_pagetable_physaddr;
  qvm86_mapped_page *guest_pgd;
  uint16_t last_mon_cs;

  /* Pointer to guest->host physical page mappings.  */
  uint32_t **monitor_physmem_index;
  /* Monitor pagetable entries for guest->host physical page mappings.  */
  uint32_t *monitor_physmem_pte;
  uint32_t monitor_physmem_pte_physaddr;
  /* Monitor pagetable entries to map the shadow pagetables into monitor
     space.  */
  uint32_t *monitor_shadow_pte;
  uint32_t monitor_shadow_pte_physaddr;

  /* The host userspace address of guest physical memory.  */
  char * user_physmem;
  unsigned long user_physmem_size;
  /* Emulator IO memory mapping.  */
  uint32_t *l1_phys_map;
  /* Emulator dirty memory flags.  */
  uint8_t *phys_ram_dirty;
  unsigned long *pages_to_flush;
  /* Satus of mapped-in host pages.  */
  qvm86_mapped_page *host_physmem_map;

  /* This points to the shared copy, and ensures the code doesn't disappear
     when we switch to monitor space.  */
  qvm86_monitorproc monitor_run;

  /* This is the location of the monitor space.  It may be moved around.  */
  uint32_t monitor_base;

  /* maddr_* are offsets from monitor_base.  */
  uint32_t maddr_handleint;
  uint32_t maddr_handlefault;
  uint32_t maddr_start_guest;

  uint32_t maddr_stubs; /* monitor_stubs */
  uint32_t maddr_gdt; /* monitor_gdt */
  uint32_t maddr_idt; /* monitor_idt */
  uint32_t maddr_stack; /* monitor_stack */
  uint32_t maddr_tss; /* monitor_tss */
  /* The monitor pde for mapping host space while switching.  */
  uint32_t maddr_switch_ptr;

  /* This will be nonzero if a fatal error occured.  Usually due to
     something not yet implemented.  Ideally this would never happen.  */
  int die;

  /* This pointer is exclusively for use by specific host code. */
  void *host_opaque;
} qvm86_state;

void qvm86_init (qvm86_state * qs, struct kqemu_init * init);
int qvm86_exec (qvm86_state * qs, struct kqemu_cpu_state * hostenv);
void qvm86_init_device (qvm86_state *);
void qvm86_cleanup_device (qvm86_state *);
/* Give other tasks a chance to run.  Returns nonzero if this task has
   signals pending.  */
int host_do_sched (qvm86_state *);
/* Allocate a contiguous block of physical memory.  The returned block must
   be suitable for use with host_phys_addr.  */
void * host_alloc_pages (int);
/* Allocate a block of host memory.  This memory is used for internal memory
   structures, and does not need to be locked in a specifric location.  */
void * host_alloc_zeroed (unsigned long);
/* Free a memory allocated with host_alloc_zeroed.  */
void host_free (void *);
/* Free memory allocated with host_alloc_pages.  */
void host_free_pages (void *, int);
/* Lock a page of user memory into kernel space.  */
void host_map_user (unsigned long, qvm86_mapped_page *);
/* Unlock a page of user memory.  */
void host_unmap_user (qvm86_mapped_page *, int);
/* Mark a page as dirty.  */
void host_mark_dirty (qvm86_mapped_page *);
host_physaddr_t host_phys_addr (void *);

/* in switch.S */
void asmlinkage qvm86_monitor_run(qvm86_monitor_env *);

/* Monitor-side routines.  */
void qvm86_monitor_return (void);
void qvm86_monitor_entry (void);
void qvm86_start_guest (void);
void qvm86_monitor_switch (void);
void qvm86_monitor_inthandler (void);
void qvm86_monitor_faulthandler (void);
